class MERBudgetCostBalance(PWReportPage):
	def init(self):
		super(MERBudgetCostBalance, self).init()
		self.project				= None
		self.withObjective			= True
		self.contentMode			= 0
		self.approvedBudgetColor	= None
		self.plannedTotalCostColor	= None
		self.actualTotalCostColor	= None
		self.iterIndex				= 0
		self.iterPhase				= None
		return self

	def showPlannedValues(self):
		return self.contentMode == 0 or self.contentMode == 2
		
	def showActualValues(self):
		return self.contentMode == 1 or self.contentMode == 2

	def showPlannedAndActualValues(self):
		return self.contentMode == 2
		
	def showOnlyPlannedValues(self):
		return self.contentMode == 0

	def showOnlyActualValues(self):
		return self.contentMode == 1

	def showObjective(self):
		return self.project.itemDescription()

	def topPhases(self):
		activities = self.project.visibleSubActivitiesArray()
		startDateDescriptor = NSSortDescriptor.alloc().initWithKey_ascending_("plannedStartDate", True)
		if activities:
			return activities.sortedArrayUsingDescriptors_([startDateDescriptor])
		else:
			return NSArray.array()
			
	def rowClass(self):
		if self.iterIndex%2:
			return "even"

	def totalRowClass(self):
		if len(self.topPhases())%2:
			return "even"
		
	def dateFormat(self):
		return "medium/short"


	def cellClassForAmount(self, amount):
		if amount < 0:
			return "negativeBalance"
		else:
			return "positiveBalance"

	def plannedBalanceCellClass(self):
		return self.cellClassForAmount(self.iterPhase.budgetPlannedCostBalance())

	def actualBalanceCellClass(self):
		return self.cellClassForAmount(self.iterPhase.budgetActualCostBalance())

	def plannedTotalBalanceCellClass(self):
		return self.cellClassForAmount(self.project.budgetPlannedCostBalance())

	def actualTotalBalanceCellClass(self):
		return self.cellClassForAmount(self.project.budgetActualCostBalance())

	def xTicks(self):
		xTicks = "\"xTicks\": [ "
		phases = self.topPhases()
		
		for i in range(len(phases)):
			phase = phases[i]
			if i<len(phases)-1:
				separator = ","
			else:
				separator = ""
			if self.withActivityLabels:
				titel = phase.title().replace("\"", "\\\"")
				xTicks += "{v:%d, label:\"%s\"}%s\n" % (i, titel, separator)
			else:
				xTicks += "{v:%d, label:\"%s\"}%s\n" % (i, " ", separator)
		xTicks = xTicks + "]"
		return xTicks
	
	def dataSetStringForKey(self, key):
		dataSetString = "\"%s\",[" % (key)
		phases = self.topPhases()
		for i in range(len(phases)):
			phase = phases[i]
			if i<len(phases)-1:
				separator = ","
			else:
				separator = ""
			value = phase.valueForKey_(key)
			if not value:
				value = 0
			dataSetString += "[%d, %s]%s" % (i, value, separator)
		dataSetString += "]"
		return dataSetString

	def budgetDataSetString(self):
		return self.dataSetStringForKey("approvedBudgetAbsolute")
		
	def plannedCostDataSetString(self):
		return self.dataSetStringForKey("plannedCost")

	def actualCostDataSetString(self):
		return self.dataSetStringForKey("actualCost")

	def approvedBudgetColorString(self):
		return self.approvedBudgetColor.webColor()

	def plannedTotalCostColorString(self):
		return self.plannedTotalCostColor.webColor()

	def actualTotalCostColorString(self):
		return self.actualTotalCostColor.webColor()
